package com.kob.backend.consumer.utils;

import com.alibaba.fastjson.JSONObject;
import com.kob.backend.consumer.WebSocketServer;
import com.kob.backend.pojo.Bot;
import com.kob.backend.pojo.Record;
import com.kob.backend.pojo.User;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.concurrent.locks.ReentrantLock;

/**
 * @author zfp
 */
public class Game extends Thread {
    private final Integer rows;
    private final Integer cols;
    private final Integer inner_walls_count;
    private final int[][] g;

    private final Player playerA, playerB;
    private final static int[] dx = {-1, 0, 1, 0};
    private final static int[] dy = {0, 1, 0, -1};

    private Integer nextStepA = null; // 玩家A的下一步操作，没有操作设置为null
    private Integer nextStepB = null;

    private ReentrantLock lock = new ReentrantLock(); // 多线程之间读写操作需要加锁
    private String status = "playing"; // playing -> finished
    private String loser = "";// "all": 平局 "A": A输 "B": B输

    private final static String addBotUrl = "http://127.0.0.1:4002/bot/add/";

    public Game(Integer rows, Integer cols, Integer inner_walls_count, Integer idA, Integer idB, Bot bot_a,Bot bot_b) {
        this.rows = rows;
        this.cols = cols;
        this.inner_walls_count = inner_walls_count;
        this.g = new int[rows][cols];
        Integer bot_a_id = -1,bot_b_id = -1;
        String bot_a_code = "",bot_b_code = "";
        if (bot_a != null) {
            bot_a_id = bot_a.getId();
            bot_a_code = bot_a.getContent();
        }

        if (bot_b != null) {
            bot_b_id = bot_b.getId();
            bot_b_code = bot_b.getContent();
        }

        this.playerA = new Player(idA, bot_a_id,bot_a_code,rows - 2, 1, new ArrayList<>());
        this.playerB = new Player(idB,bot_b_id,bot_b_code, 1, cols - 2, new ArrayList<>());
    }

    public int[][] getG() {
        return g;
    }

    private boolean check_connectivity(int sx, int sy, int tx, int ty) {
        if (sx == tx && sy == ty) {
            return true;
        }
        g[sx][sy] = 1;
        for (int i = 0; i < 4; i++) {
            int x = sx + dx[i], y = sy + dy[i];
            if (x >= 0 && x < this.rows && y >= 0 && y < this.cols && g[x][y] == 0) {
                if (check_connectivity(x, y, tx, ty)) {
                    g[sx][sy] = 0;
                    return true;
                }
            }
        }
        g[sx][sy] = 0;
        return false;
    }

    // 绘制地图是否成功
    private boolean draw() {
        for (int i = 0; i < this.rows; i++) {
            for (int j = 0; j < this.cols; j++) {
                // 0表示空地，1表示障碍物
                g[i][j] = 0;
            }
        }

        // 四周加上障碍物

        for (int r = 0; r < this.rows; r++) {
            g[r][0] = g[r][this.cols - 1] = 1;
        }

        for (int c = 0; c < this.cols; c++) {
            g[0][c] = g[this.rows - 1][c] = 1;
        }

        // 创建随机障碍物

        Random random = new Random();

        for (int i = 0; i < this.inner_walls_count / 2; i++) {
            for (int j = 0; j < 1000; j++) {
                int r = random.nextInt(this.rows);
                int c = random.nextInt(this.cols);
                // 判断对称的位置是否有障碍物
                if (g[r][c] == 1 || g[this.rows - 1 - r][this.cols - 1 - c] == 1) {
                    continue;
                }
                // 不要生成到蛇初始的位置上
                if (r == this.rows - 2 && c == 1 || r == 1 && c == this.cols - 2) {
                    continue;
                }
                // 生成的位置没有问题
                g[r][c] = g[this.rows - 1 - r][this.cols - 1 - c] = 1;
                break;
            }
        }
        return this.check_connectivity(this.rows - 2, 1, 1, this.cols - 2);
    }

    public void createMap() {
        // 最多随机1000次,判断能够生成有效的地图
        for (int i = 0; i < 1000; i++) {
            if (draw()) {
                break;
            }
        }
    }

    public Player getPlayerA() {
        return playerA;
    }

    public Player getPlayerB() {
        return playerB;
    }

    // 多个client可能会同时进行写操作，需要加锁
    public void setNextStepA(Integer nextStepA) {
        lock.lock();
        try {
            this.nextStepA = nextStepA;
        } finally {
            lock.unlock();
        }
    }

    public void setNextStepB(Integer nextStepB) {
        lock.lock();
        try {
            this.nextStepB = nextStepB;
        } finally {
            lock.unlock();
        }
    }

    private String getInput(Player player) {
        Player me,you;
        if (playerA.getId().equals(player.getId())) {
            me = playerA;
            you = playerB;
        }else {
            me = playerB;
            you = playerA;
        }
        return getMapString() + "#" +
                me.getSx() + "#" +
                me.getSy() + "#(" +
                me.getStepsString() + ")#" +
                you.getSx() + "#" +
                you.getSy() + "#(" +
                you.getStepsString() + ")";
    }

    // 是否需要使用bot代码
    private void sendBotCode(Player player) {
        if (player.getBotId().equals(-1)) { // 用户出战直接返回
            return;
        }

        MultiValueMap<String,String> data = new LinkedMultiValueMap<>();
        data.add("user_id",player.getId().toString());
        data.add("bot_code",player.getBotCode());
        data.add("input",getInput(player));

        // 向BotRunningSystem发送信息
        WebSocketServer.restTemplate.postForObject(addBotUrl,data,String.class);
    }

    public boolean nextStep() {
        try {
            // 保证前端不会遗漏操作，前端是每200ms移动一个格子
            Thread.sleep(200);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        sendBotCode(playerA);
        sendBotCode(playerB);


        // 10s倒计时操作，10s内没有输入判输
        for (int i = 0; i < 100; i++) {
            try {
                Thread.sleep(100);
                lock.lock();
                try {
                    // 两名用户都进行了操作
                    if (nextStepA != null && nextStepB != null) {
                        // 记录用户的操作
                        playerA.getSteps().add(nextStepA);
                        playerB.getSteps().add(nextStepB);
                        return true;
                    }
                } finally {
                    lock.unlock();
                }
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        return false;
    }

    private void sendAllMessage(String message) { // 向两个client传递信息
        if (WebSocketServer.users.get(playerA.getId()) != null)
            WebSocketServer.users.get(playerA.getId()).sendMessage(message);
        if (WebSocketServer.users.get(playerB.getId()) != null)
            WebSocketServer.users.get(playerB.getId()).sendMessage(message);

    }

    private String getMapString() {
        StringBuilder res = new StringBuilder();
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                res.append(g[i][j]);
            }
        }
        return res.toString();
    }

    // 更新用户的积分
    private void updateUserRating(Player player,Integer rating) {
        User user = WebSocketServer.userMapper.selectById(player.getId());
        user.setRating(rating);
        WebSocketServer.userMapper.updateById(user);
    }


    // 将对局记录存入数据库
    private void saveRecords() {
        Integer ratingA = WebSocketServer.userMapper.selectById(playerA.getId()).getRating();
        Integer ratingB = WebSocketServer.userMapper.selectById(playerB.getId()).getRating();

        if("A".equals(loser)) {
            ratingA -= 2; // 输者减两分
            ratingB += 5; // 赢者加5分
        }else if ("B".equals(loser)) {
            ratingA += 5;
            ratingB -= 2;
        }

        // 更新两个用户的积分
        updateUserRating(playerA,ratingA);
        updateUserRating(playerB,ratingB);


        Record record = new Record(
                null,
                playerA.getId(),
                playerA.getSx(),
                playerA.getSy(),
                playerB.getId(),
                playerB.getSx(),
                playerB.getSy(),
                playerA.getStepsString(),
                playerB.getStepsString(),
                getMapString(),
                loser,
                new Date()
        );
        WebSocketServer.recordMapper.insert(record);
    }



    private void sendResult() { // 向两个client发送结果
        JSONObject resp = new JSONObject();
        resp.put("event", "result");
        resp.put("loser", loser);
        saveRecords();
        sendAllMessage(resp.toJSONString());
    }

    private boolean check_valid(List<Cell> cellsA,List<Cell> cellsB) {
        int n = cellsA.size();
        Cell c = cellsA.get(n-1);

        if (g[c.getX()][c.getY()] == 1) {return false;}

        for (int i = 0; i < n-1; i++) {
            if (cellsA.get(i).getX() == c.getX() && cellsA.get(i).getY() == c.getY()) {
                return false;
            }
        }

        for (int i = 0; i < n-1; i++) {
            if (cellsB.get(i).getX() == c.getX() && cellsB.get(i).getY() == c.getY()) {
                return false;
            }
        }
        return true;
    }

    private void judge() { // 判断两名玩家的下一步是否合法
        List<Cell> cellsA = playerA.getCells();
        List<Cell> cellsB = playerB.getCells();

        boolean valid_A = check_valid(cellsA,cellsB);
        boolean valid_B = check_valid(cellsB,cellsA);

        if (!valid_A || !valid_B) {
            status = "finished";
            if (!valid_A && !valid_B) {
                loser = "all";
            }else if (!valid_A) {
                loser = "A";
            }else {
                loser = "B";
            }
        }

    }

    private void sendMove() { // 向两名玩家传递蛇的移动信息
        lock.lock();
        try {
            JSONObject resp = new JSONObject();
            resp.put("event", "move");
            resp.put("a_direction", nextStepA);
            resp.put("b_direction", nextStepB);
            sendAllMessage(resp.toJSONString());
            nextStepA = nextStepB = null;
        } finally {
            lock.unlock();
        }

    }

    @Override
    public void run() {
        for (int i = 0; i < 1000; i++) {
            if (nextStep()) {
                judge();
                if ("playing".equals(status)) {
                    sendMove();
                } else {
                    sendResult();
                    break;
                }
            } else {
                status = "finished";
                lock.lock();
                try {
                    if (nextStepA == null && nextStepB == null) {
                        loser = "all";
                    } else if (nextStepA == null) {
                        loser = "A";
                    } else {
                        loser = "B";
                    }
                } finally {
                    lock.unlock();
                }
                sendResult();
                break;
            }
        }
    }


}
